#include "torsmo.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/sysinfo.h>

static int meminfo_updated;
static int sysinfo_updated;
static int stat_updated;
static int net_updated;
static struct sysinfo s_info;

void prepare_update() {
  sysinfo_updated = 0;
  meminfo_updated = 0;
  stat_updated = 0;
  net_updated = 0;
}

static void update_sysinfo() {
  if (sysinfo_updated) return;
  sysinfo_updated = 1;

  sysinfo(&s_info);
}

double get_uptime() {
  double t = 0;

  /* prefers sysinfo() for uptime, I don't really know which one is better
   * (=faster?) */

#ifdef USE_PROC_UPTIME
  static int rep;
  FILE *fp = open_file("/proc/uptime", &rep);
  if(!fp) return 0;
  fscanf(fp, "%lf", &t);
  fclose(fp);
#else
  update_sysinfo();
  t = (double) s_info.uptime;
#endif

  return t;
}

/* these things are also in sysinfo except Buffers:, that's why I'm reading
 * them from proc */
void update_meminfo() {
  static int rep;
  unsigned int a;
  char buf[256];
  FILE *fp;

  if (meminfo_updated) return;
  meminfo_updated = 1;

  mem = memmax = swap = swapmax = bufmem = 0;

  fp = open_file("/proc/meminfo", &rep);
  if (!fp) return;

  while (!feof(fp)) {
    if (fgets(buf, 255, fp) == NULL) break;

    if (strncmp(buf, "MemTotal:", 9) == 0) {
      sscanf(buf, "%*s %u", &memmax);
    }
    else if (strncmp(buf, "MemFree:", 8) == 0) {
      sscanf(buf, "%*s %u", &mem);
    }
    else if (strncmp(buf, "SwapTotal:", 10) == 0) {
      sscanf(buf, "%*s %u", &swapmax);
    }
    else if (strncmp(buf, "SwapFree:", 9) == 0) {
      sscanf(buf, "%*s %u", &swap);
    }
    else if (strncmp(buf, "Buffers:", 8) == 0) {
      sscanf(buf, "%*s %u", &a);
      bufmem += a;
    }
    else if (strncmp(buf, "Cached:", 7) == 0) {
      sscanf(buf, "%*s %u", &a);
      bufmem += a;
    }
  }

  fclose(fp);

  mem = memmax - mem;
  swap = swapmax - swap;

  if (no_buffers) mem -= bufmem;
}

static double last_net_time;

void update_net_stats() {
  static int rep;
  unsigned int i;
  char buf[256];
  FILE *fp;
  double t, delta;

  if (net_updated) return;
  net_updated = 1;

  /* get delta */
  t = get_time();
  delta = t - last_net_time;
  if (delta <= 0.0001) return;
  last_net_time = t;

  /* open file and ignore first two lines */
  fp = open_file("/proc/net/dev", &rep);
  if (!fp) return;

  fgets(buf, 255, fp); /* garbage */
  fgets(buf, 255, fp); /* garbage (field names) */

  /* read each interface */
  for (i=0; i<16; i++) {
    struct net_stat *ns;
    char *s, *p;
    long long r, t, last_recv, last_trans;

    if (fgets(buf, 255, fp) == NULL) break;
    p = buf;
    while (isspace((int) *p)) p++;

    s = p;

    while (*p && *p != ':') p++;
    if (*p == '\0') continue;
    *p = '\0';
    p++;

    ns = get_net_stat(s);
    last_recv = ns->recv;
    last_trans = ns->trans;

    sscanf(p,
    /* bytes packets errs drop fifo frame compressed multicast|bytes ... */
       "%Ld  %*d     %*d  %*d  %*d  %*d   %*d        %*d       %Ld",
        &r,                                                    &t);

    /* if recv or trans is less than last time, an overflow happened */

    if (r < ns->last_read_recv)
      ns->recv += ((long long) 4294967295U - ns->last_read_recv) + r;
    else
      ns->recv += (r - ns->last_read_recv);
    ns->last_read_recv = r;

    if (t < ns->last_read_trans)
      ns->trans += (4294967295U - ns->last_read_trans) + t;
    else
      ns->trans += (t - ns->last_read_trans);
    ns->last_read_trans = t;

    /* calculate speeds */
    ns->recv_speed = (ns->recv - last_recv) / delta;
    ns->trans_speed = (ns->trans - last_trans) / delta;
  }

  fclose(fp);
}

int get_total_processes() {
  update_sysinfo();
  return s_info.procs;
}

static int running_processes;
static unsigned int cpu_user, cpu_system, cpu_nice;

static void update_stat() {
  static int rep;
  char buf[256];
  FILE *fp;

  if (stat_updated) return;
  stat_updated = 1;

  fp = open_file("/proc/stat", &rep);
  if (!fp) return;

  while (!feof(fp)) {
    if (fgets(buf, 255, fp) == NULL)
      break;

    if (strncmp(buf, "procs_running ", 14) == 0) {
      sscanf(buf, "%*s %d", &running_processes);
    }
    else if (strncmp(buf, "cpu ", 4) == 0) {
      sscanf(buf, "%*s %u %u %u", &cpu_user, &cpu_nice, &cpu_system);
    }
  }

  fclose(fp);
}

int get_running_processes() {
  update_stat();
  return running_processes;
}

static double last_cpu_sum;
static double last_cpu_calc_time;

void update_cpu_usage() {
  double t, delta;

  t = get_time();
  delta = t - last_cpu_calc_time;

  if(delta <= 0.05) return;

  last_cpu_calc_time = t;

  update_stat();

  cpu_usage = (cpu_user+cpu_nice+cpu_system - last_cpu_sum) / delta / 100.0;
  last_cpu_sum = cpu_user+cpu_nice+cpu_system;
}

static const char *default_i2c_dev;

static int no_dots(const struct dirent *d) {
  if(d->d_name[0] == '.') return 0;
  return 1;
}

#define I2C_DIRECTORY "/sys/bus/i2c/devices/"

double get_i2c_info(const char *dev, const char *type, int n) {
  char path[256];
  FILE *fp;
  int val = 0;

  /* if i2c device is NULL or *, get first */
  if (dev == NULL || strcmp(dev, "*") == 0) {
    if (default_i2c_dev == NULL) {
      struct dirent **namelist;
      int i, n;

      n = scandir(I2C_DIRECTORY, &namelist, no_dots, alphasort);
      if (n < 0) {
        static int rep;
        if (!rep) {
          ERR("scandir for %s: %s", I2C_DIRECTORY, strerror(errno));
          rep = 1;
        }
        return 0;
      }
      else {
        if (n == 0) {
          /* no sensors, oh dear. */
          return 0;
        }

        /* get default i2c device and free others (leaks a bit but this is
         * done only once per run) */
        default_i2c_dev = namelist[0]->d_name;

        for (i=1; i<n; i++)
          free(namelist[i]);
        free(namelist);
      }
    }

    dev = default_i2c_dev;
  }

  /* change vol to in */
  if (strcmp(type, "vol") == 0)
    type = "in";

  snprintf(path, 255, "/sys/bus/i2c/devices/%s/%s%d_input", dev, type, n);

  /* open file, read value, close file */
  fp = open_file(path, 0);
  if (!fp) return 0;
  fscanf(fp, "%d", &val);
  fclose(fp);

  /* divide voltage and temperature by 1000 */
  if (strcmp(type, "in") == 0)
    return val / 1000.0;
  if (strcmp(type, "temp") == 0)
    return val / 1000.0;
  else
    return val;
}
