/* torsmo, a system monitor
 *
 * This program is licensed under BSD license, read COPYING
 *
 * see torsmorc.sample for configuration
 */

#include "torsmo.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>
#include <sys/utsname.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>

/* alignments */
enum {
  TOP_LEFT = 1,
  TOP_RIGHT,
  BOTTOM_LEFT,
  BOTTOM_RIGHT
};

/* default stuff */

/* default config file, undef if you don't want to use this at all */
#define CONFIG_FILE "$HOME/.torsmorc"

/* set to 1 if you want all text to be in uppercase */
static unsigned int stuff_in_upper_case = 0;

/* Position on the screen */
static int text_alignment = BOTTOM_LEFT;
static int gap_x = 5;
static int gap_y = 5;

/* Font used */
static const char *font = "6x10";

/* Update interval */
static double update_interval = 10.0;

/* fork? */
static int fork_to_background = 0;

/* border */
static int draw_borders = 0;
static int stippled_borders = 0;

static int draw_shades = 1;

static long default_fg_color;
static long default_bg_color;

/* no buffers in used memory? */
int no_buffers = 1;

/* Text that is shown */
static char original_text[] =
"$nodename - $sysname $kernel on $machine\n"
"$hr\n"
"${color grey}Uptime$color $uptime\n"
"${color grey}RAM Usage:$color $mem/$memmax - $memperc% ${membar 4}\n"
"${color grey}Swap Usage:$color $swap/$swapmax - $swapperc% ${swapbar 4}\n"
"${color grey}CPU Usage:$color $cpu% ${cpubar 4}\n"
"${color grey}Processes:$color $processes  ${color grey}Running:$color $running_processes\n"
"$hr\n"
"${color grey}Networking:\n"
" Up:$color ${upspeed eth0} k/s${color grey} - Down:$color ${downspeed eth0} k/s\n"
"${color grey}Temperatures:\n"
" CPU:$color ${i2c temp 1}C${color grey} - MB:$color ${i2c temp 2}C\n"
;

static char *text = original_text;

static Display *display;
static int screen;
static GC gc;
static XFontStruct *font_info;
static Window win;
static int display_width, display_height;

/* formatted text to render on screen, generated in generate_text(),
 * drawn in draw_stuff() */

#define TEXT_BUFFER_SIZE 2048

static char text_buffer[TEXT_BUFFER_SIZE];

/* special stuff in text_buffer */

#define SPECIAL_CHAR '\x01'

enum {
  HORIZONTAL_LINE,
  STIPPLED_HR,
  BAR,
  FG,
  BG,
};

static struct special_t {
  int type;
  int height;
  long arg;
} specials[128];

static int special_count;

static void new_bar(char *buf, const char *arg, int usage) {
  buf[0] = SPECIAL_CHAR;
  buf[1] = '\0';

  specials[special_count].type = BAR;
  specials[special_count].height = arg ? atoi(arg) : 4;
  specials[special_count].arg = (usage > 255) ? 255 : ((usage < 0) ? 0 : usage);
  special_count++;
}

static void new_hr(char *buf, const char *arg) {
  buf[0] = SPECIAL_CHAR;
  buf[1] = '\0';

  specials[special_count].type = HORIZONTAL_LINE;
  specials[special_count].height = arg ? atoi(arg) : 1;
  special_count++;
}

static void new_stippled_hr(char *buf, int a, int b) {
  buf[0] = SPECIAL_CHAR;
  buf[1] = '\0';

  specials[special_count].type = STIPPLED_HR;
  specials[special_count].height = b;
  specials[special_count].arg = a;
  special_count++;
}

static unsigned long get_color(const char *name) {
  XColor color;
  XWindowAttributes attrs;

  XGetWindowAttributes(display, win, &attrs);
  color.pixel = 0;
  if (!XParseColor(display, attrs.colormap, name, &color))
    ERR("can't parse X color '%s'", name);
  else if (!XAllocColor(display, attrs.colormap, &color))
    ERR("can't allocate X color '%s'", name);

  return color.pixel;
}

static void new_fg(char *buf, const char *arg) {
  buf[0] = SPECIAL_CHAR;
  buf[1] = '\0';

  specials[special_count].type = FG;
  specials[special_count].arg = arg ? get_color(arg) : default_fg_color;
  special_count++;
}

static void new_bg(char *buf, const char *arg) {
  buf[0] = SPECIAL_CHAR;
  buf[1] = '\0';

  specials[special_count].type = BG;
  specials[special_count].arg = arg ? get_color(arg) : default_bg_color;
  special_count++;
}

/* quite boring functions */

double get_time() {
  struct timeval tv;
  gettimeofday(&tv, 0);
  return tv.tv_sec + tv.tv_usec / 1000000.0;
}

FILE *open_file(const char *file, int *reported) {
  FILE *fp = fopen(file, "r");
  if (!fp) {
    if (!reported || *reported == 0) {
      ERR("can't open %s: %s", file, strerror(errno));
      if (reported) *reported = 1;
    }
    return 0;
  }

  return fp;
}

static void for_each_line(char *b, void (*f)(char *)) {
  char *ps, *pe;

  for (ps=b, pe=b; *pe; pe++) {
    if (*pe == '\n') {
      *pe = '\0';
      f(ps);
      *pe = '\n';
      ps = pe+1;
    }
  }

  if (ps < pe) f(ps);
}

static void convert_escapes(char *buf) {
  char *p = buf, *s = buf;

  while (*s) {
    if(*s == '\\') {
      s++;
      if(*s == 'n') *p++ = '\n';
      else if(*s == '\\') *p++ = '\\';
      s++;
    }
    *p++ = *s++;
  }
  *p = '\0';
}

/* memory info */

unsigned int mem=0, memmax=0, swap=0, swapmax=0, bufmem=0;

/* cpu usage stuff */

double cpu_usage = 0.0;

/* network interface stuff */

static struct net_stat netstats[16];

struct net_stat *get_net_stat(const char *dev) {
  unsigned int i;

  /* find interface stat */
  for (i=0; i<16; i++) {
    if (netstats[i].dev && strcmp(netstats[i].dev, dev) == 0)
      return &netstats[i];
  }

  /* wasn't found? add it */
  if (i == 16) {
    for (i=0; i<16; i++) {
      if (netstats[i].dev == 0) {
        netstats[i].dev = strdup(dev);
        return &netstats[i];
      }
    }
  }

  CRIT_ERR("too many interfaces used (limit is 16)");
  return 0;
}

static void human_readable(long long a, char *buf) {
  if (a >= 1024*1024*1024)
    snprintf(buf, 255, "%.2fG", (a/1024/1024)/1024.0);
  else if (a >= 1024*1024)
    snprintf(buf, 255, "%.1fM", (a/1024)/1024.0);
  else if (a >= 1024)
    snprintf(buf, 255, "%Ldk", a/1024);
  else
    snprintf(buf, 255, "%Ld", a);
}

static struct utsname uname_s;

static void update_uname() {
  uname(&uname_s);
}

static void get_special_variable(const char *s, const char *arg, char *buf) {
#define VAR(a) else if (strcasecmp(s, #a) == 0)

  if (s[0] == '#') {
    new_fg(buf, s);
  }
  VAR(cpu) {
    update_cpu_usage();
    snprintf(buf, 255, "%d", (int) (cpu_usage*100.0));
  }
  VAR(cpubar) {
    update_cpu_usage();
    new_bar(buf, arg, (int) (cpu_usage*255.0));
  }
  VAR(color) {
    new_fg(buf, arg);
  }
  VAR(debugbar) {
    new_bar(buf, arg, atoi(arg+2));
  }
  VAR(downspeed) {
    update_net_stats();
    snprintf(buf, 255, "%d", (int) (get_net_stat(arg)->recv_speed/1024));
  }
  VAR(downspeedf) {
    update_net_stats();
    snprintf(buf, 255, "%.1f", get_net_stat(arg)->recv_speed/1024.0);
  }
  VAR(hr) {
    new_hr(buf, arg);
  }
  VAR(i2c) {
    char buf1[64], buf2[64];
    int n;
    double r;

    if(sscanf(arg, "%63s %63s %d", buf1, buf2, &n) != 3) {
      /* if scanf couldn't read three values, read type and num and use
       * default device */
      sscanf(arg, "%63s %d", buf2, &n);
      buf1[0] = '*';
      buf1[1] = '\0';
    }
    buf1[63] = '\0';
    buf2[63] = '\0';

    r = get_i2c_info(buf1, buf2, n);

    /* put integer if it's fan (RPM), else float (temp or vol) */
    if (strcmp(buf2, "fan") == 0)
      snprintf(buf, 255, "%d", (int) r);
    else
      snprintf(buf, 255, "%.1f", r);
  }
  VAR(kernel) {
    strcpy(buf, uname_s.release);
  }
  VAR(machine) {
    strcpy(buf, uname_s.machine);
  }
  VAR(mem) {
    update_meminfo();
    human_readable(mem*1024, buf);
  }
  VAR(memmax) {
    update_meminfo();
    human_readable(memmax*1024, buf);
  }
  VAR(memperc) {
    update_meminfo();
    if (memmax)
      snprintf(buf, 255, "%d", (mem*100) / (memmax));
  }
  VAR(membar) {
    update_meminfo();
    new_bar(buf, arg, memmax ? (mem*255) / (memmax) : 0);
  }
  VAR(nodename) {
    strcpy(buf, uname_s.nodename);
  }
  VAR(processes) {
    snprintf(buf, 255, "%d", get_total_processes());
  }
  VAR(running_processes) {
    snprintf(buf, 255, "%d", get_running_processes());
  }
  VAR(shadecolor) {
    new_bg(buf, arg);
  }
  VAR(stippled_hr) {
    int a = stippled_borders, b = 1;
    if(arg) {
      if(sscanf(arg, "%d %d", &a, &b) != 2)
        sscanf(arg, "%d", &b);
    }
    new_stippled_hr(buf, a, b);
  }
  VAR(swap) {
    update_meminfo();
    human_readable(swap*1024, buf);
  }
  VAR(swapmax) {
    update_meminfo();
    human_readable(swapmax*1024, buf);
  }
  VAR(swapperc) {
    update_meminfo();
    if (swapmax)
      snprintf(buf, 255, "%d", (swap*100) / (swapmax));
  }
  VAR(swapbar) {
    update_meminfo();
    new_bar(buf, arg, swapmax ? (swap*255) / (swapmax) : 0);
  }
  VAR(sysname) {
    strcpy(buf, uname_s.sysname);
  }
  VAR(temp1) {
    snprintf(buf, 255, "%.1f", get_i2c_info(0, "temp", 1));
  }
  VAR(temp2) {
    snprintf(buf, 255, "%.1f", get_i2c_info(0, "temp", 2));
  }
  VAR(time) {
    time_t t = time(NULL);
    struct tm *tm = localtime(&t);
    strftime(buf, 255, arg ? arg : "%F %T", tm);
  }
  VAR(totaldown) {
    update_net_stats();
    if (arg)
      human_readable(get_net_stat(arg)->recv, buf);
  }
  VAR(totalup) {
    update_net_stats();
    if (arg)
      human_readable(get_net_stat(arg)->trans, buf);
  }
  VAR(updates) {
    static int total_updates = 0;
    snprintf(buf, 255, "%d", total_updates++);
  }
  VAR(upspeed) {
    update_net_stats();
    snprintf(buf, 255, "%d", (int) (get_net_stat(arg)->trans_speed/1024));
  }
  VAR(upspeedf) {
    update_net_stats();
    snprintf(buf, 255, "%.1f", get_net_stat(arg)->trans_speed/1024.0);
  }
  VAR(uptime) {
    double t = get_uptime();

    if(t >= 24*60*60)
      snprintf(buf, 255, "%dd %dh %dmin", (int) (t/60.0/60.0/24.0),
          (int) (t/60.0/60.0) % 24,
          (int) (t/60.0) % 60);
    else if(t >= 60*60)
      snprintf(buf, 255, "%dh %dmin",
          (int) (t/60.0/60.0) % 24,
          (int) (t/60.0) % 60);
    else
      snprintf(buf, 255, "%dmin", (int) (t/60.0));
  }
  else {
    buf[0] = '$';
    buf[1] = '{';
    strcpy(buf+2, s);
    strcat(buf, "}");
  }
}

static void variable_substitute(const char *s, char *dest, unsigned int n) {
  while (*s && n > 1) {
    if (*s == '$') {
      s++;
      if (*s != '$') {
        char buf[256], buf2[256];
        const char *a, *var;
        unsigned int len;

        /* variable is either $foo or ${foo} */
        if (*s == '{') {
          s++;
          a = s;
          while(*s && *s != '}') s++;
        }
        else {
          a = s;
          if (*s == '#') s++;
          while(*s && (isalnum((int) *s) || *s=='_')) s++;
        }

        /* copy variable to buffer and look it up */
        len = (s-a > 255) ? 255 : (s-a);
        strncpy(buf, a, len);
        buf[len] = '\0';

        if (*s == '}') s++;

        var = getenv(buf);

        /* if variable wasn't found from environment, use some special */
        if (!var) {
          char *arg = 0;

          if (strchr(buf, ' ')) {
            arg = strchr(buf, ' ');
            *arg = '\0';
            arg++;
            while (isspace((int) *arg)) arg++;
            if (!*arg) arg = 0;
          }

          get_special_variable(buf, arg, buf2);
          var = buf2;
        }

        /* add var to dest */
        len = strlen(var);
        if (len >= n) len = n-1;
        strncpy(dest, var, len);
        dest += len;
        n -= len;
        continue;
      }
    }

    *dest++ = *s++;
    n--;
  }

  *dest = '\0';
}

static double last_update_time;

static void generate_text() {
  unsigned int i;

  /* set net speeds to zero in case device was removed and doesn't get
   * updated */
  for (i=0; i<16; i++) {
    if (netstats[i].dev) {
      netstats[i].recv_speed = netstats[i].trans_speed = 0.0;
    }
  }

  special_count = 0;

  prepare_update();

  variable_substitute(text, text_buffer, TEXT_BUFFER_SIZE);

  last_update_time = get_time();

  if (stuff_in_upper_case) {
    char *p;

    p = text_buffer;
    while (*p) {
      *p = toupper(*p);
      p++;
    }
  }
}

/* text size */

static int text_width=200, text_height=200;

static inline int get_string_width(const char *s) {
  if (!*s) return 0;

  return XTextWidth(font_info, s, strlen(s));
}

static void text_size_updater(char *s) {
  int w = 0;
  char *p;

  /* get string widths and skip specials */
  p = s;
  while (*p) {
    if (*p == SPECIAL_CHAR) {
      *p = '\0';
      w += get_string_width(s);
      *p = SPECIAL_CHAR;
      s = p+1;
    }
    p++;
  }

  w += get_string_width(s);

  if (w > text_width) text_width = w;

  text_height += font_info->max_bounds.ascent + 2 +
                 font_info->max_bounds.descent;
}

/* drawing stuff */

static int text_start_x, text_start_y;
static int x, y;
static int specials_used;
static int draw_mode; /* FG or BG */

void set_foreground_color(unsigned long c) {
  XSetForeground(display, gc, c);
}

static void draw_string(char *s) {
  if (s[0] == '\0') return;

  XDrawString(display, win, gc, x, y, s, strlen(s));

  x += get_string_width(s);
}

static void draw_line(char *s) {
  char *p;

  x = text_start_x;
  y += font_info->max_bounds.ascent + 2;

  /* find specials and draw stuff */
  p = s;
  while (*p) {
    if (*p == SPECIAL_CHAR) {
      int w = 0;

      /* draw string before special */
      *p = '\0';
      draw_string(s);
      *p = SPECIAL_CHAR;
      s = p+1;

      /* draw special */
      switch (specials[specials_used].type) {
      case HORIZONTAL_LINE:
        {
          int h = specials[specials_used].height;
          int mid = font_info->max_bounds.ascent / 2;
          w = text_start_x + text_width - x;

          XSetLineAttributes(display, gc, h, LineSolid, CapButt, JoinMiter);
          XDrawLine(display, win, gc, x, y-mid, x + w + 1, y-mid);
        }
        break;

      case STIPPLED_HR:
        {
          int h = specials[specials_used].height;
          int s = specials[specials_used].arg;
          int mid = font_info->max_bounds.ascent / 2;
          w = text_start_x + text_width - x;

          XSetLineAttributes(display, gc, h, LineOnOffDash, CapButt, JoinMiter);
          XSetDashes(display, gc, 0, (char[]){s, s}, 2);
          XDrawLine(display, win, gc, x, y-mid, x + w + 1, y-mid);
        }
        break;

      case BAR:
        {
          int h = specials[specials_used].height;
          int bar_usage = specials[specials_used].arg;
          int by = y - (font_info->max_bounds.ascent + h)/2 + 1;
          w = text_start_x + text_width - x;

          XSetLineAttributes(display, gc, 1, LineSolid, CapButt, JoinMiter);

          XDrawRectangle(display, win, gc, x, by, w, h);
          XFillRectangle(display, win, gc, x, by, w * bar_usage / 255, h);
        }
        break;

      case FG:
        if (draw_mode == FG)
          set_foreground_color(specials[specials_used].arg);
        break;

      case BG:
        if (draw_mode == BG)
          set_foreground_color(specials[specials_used].arg);
        break;
      }

      x += w;

      specials_used++;
    }

    p++;
  }

  draw_string(s);

  y += font_info->max_bounds.descent;
}

static void draw_text() {
  y = text_start_y;

  /* draw borders */
  if (draw_borders) {
    if(stippled_borders) {
      XSetLineAttributes(display, gc, 1, LineOnOffDash, CapButt, JoinMiter);
      XSetDashes(display, gc, 0, (char[]){stippled_borders, stippled_borders}, 2);
    }
    else {
      XSetLineAttributes(display, gc, 1, LineSolid, CapButt, JoinMiter);
    }

    XDrawRectangle(display, win, gc, text_start_x-2, text_start_y-2,
        text_width+4, text_height+4);
  }

  /* draw text */
  specials_used = 0;
  for_each_line(text_buffer, draw_line);
}

static void update_text_area() {
  /* update text size */
  text_width = 1;
  text_height = 1;
  for_each_line(text_buffer, text_size_updater);

  /* update text start position */
  switch (text_alignment) {
  case TOP_LEFT:
    text_start_x = gap_x;
    text_start_y = gap_y;
    break;

  case TOP_RIGHT:
    text_start_x = display_width - text_width - gap_x;
    text_start_y = gap_y;
    break;

  case BOTTOM_LEFT:
    text_start_x = gap_x;
    text_start_y = display_height - text_height - gap_y;
    break;

  case BOTTOM_RIGHT:
    text_start_x = display_width - text_width - gap_x;
    text_start_y = display_height - text_height - gap_y;
    break;
  }
}

static void draw_stuff() {
  XSetLineAttributes(display, gc, 1, LineSolid, CapButt, JoinMiter);

  text_start_x++;
  text_start_y++;
  if(draw_shades) {
    set_foreground_color(default_bg_color);
    draw_mode = BG;
    draw_text();
  }

  text_start_x--;
  text_start_y--;
  set_foreground_color(default_fg_color);
  draw_mode = FG;
  draw_text();
}

static int need_to_update;

static void update_text() {
  generate_text();
  XClearArea(display, win, text_start_x-4, text_start_y-4,
      text_width+8, text_height+8, True);
  need_to_update = 1;
}

/* prototype for hup_handler ;) */
static void load_config_file(const char *);
/* a pointer to a string containing whatever configfile we are using :) */
static char *cconfig = NULL;
/* handles SIGHUPs by reloading config */
static void hup_handler(int a) {
  if (cconfig) {
    load_config_file(cconfig);
    update_text();
  }
}

static void clean_up() {
  XClearArea(display, win, text_start_x-4, text_start_y-4,
      text_width+8, text_height+8, True);
  XFlush(display);

  /* it is really pointless to free() memory at the end of program but ak|ra
   * wants me to do this */
  if (text != original_text)
    free(text);
}

static void term_handler(int a) {
  clean_up();
  exit(0);
}

static void main_loop() {
  XEvent ev;
  Region region = XCreateRegion();

  while (1) {
    XFlush(display);

    /* wait for X event or timeout */

    if (!XPending(display)) {
      fd_set fdsr;
      struct timeval tv;
      int s;
      double t = update_interval - (get_time() - last_update_time);

      if (t < 0) t = 0;

      tv.tv_sec = (long) t;
      tv.tv_usec = (long) (t * 1000000) % 1000000;

      FD_ZERO(&fdsr);
      FD_SET(ConnectionNumber(display), &fdsr);

      s = select(ConnectionNumber(display) + 1, &fdsr, 0, 0, &tv);
      if (s == -1) {
        if (errno != EINTR)
          ERR("can't select(): %s", strerror(errno));
      }
      else {
        /* timeout */
        if (s == 0) {
          update_text();
        }
      }
    }

    /* handle X events */

    while (XPending(display)) {
      XNextEvent(display, &ev);

      switch (ev.type) {
      case Expose:
        /* if text is already drawn, expose isn't too interesting if it isn't
         * on our area */

        if (ev.xexpose.x > text_start_x+text_width+4 ||
            ev.xexpose.y > text_start_y+text_height+4 ||
            ev.xexpose.x+ev.xexpose.width < text_start_x-4 ||
            ev.xexpose.y+ev.xexpose.height < text_start_y-4)
          break;

        {
          XRectangle r;
          r.x = ev.xexpose.x;
          r.y = ev.xexpose.y;
          r.width = ev.xexpose.width;
          r.height = ev.xexpose.height;
          XUnionRectWithRegion(&r, region, region);
        }
        break;

      default:
        break;
      }
    }

    if (need_to_update) {
      need_to_update = 0;

      update_text_area();

      XRectangle r;
      r.x = text_start_x-4;
      r.y = text_start_y-4;
      r.width = text_width+8;
      r.height = text_height+8;;
      XUnionRectWithRegion(&r, region, region);
    }

    if (!XEmptyRegion(region)) {
      XSetRegion(display, gc, region);
      draw_stuff();
      XDestroyRegion(region);
      region = XCreateRegion();
      XFlush(display);
    }
  }
}

static int string_to_bool(const char *s) {
  if (!s) return 1;
  if (strcasecmp(s, "yes") == 0) return 1;
  if (strcasecmp(s, "true") == 0) return 1;
  if (strcasecmp(s, "1") == 0) return 1;
  return 0;
}

static void load_config_file(const char *f) {
  int line = 0;
  FILE *fp = open_file(f, 0);
  if (!fp) return;

  while (!feof(fp)) {
    char buf[256], *p, *p2, *name, *value;
    line++;
    if (fgets(buf, 256, fp) == NULL) break;

    p = buf;

    /* break at comment */
    p2 = strchr(p, '#');
    if (p2) *p2 = '\0';

    /* skip spaces */
    while (*p && isspace((int) *p)) p++;
    if (*p == '\0') continue; /* empty line */

    name = p;

    /* skip name */
    p2 = p;
    while (*p2 && !isspace((int) *p2)) p2++;
    if (*p2 != '\0') {
      *p2 = '\0'; /* break at name's end */
      p2++;
    }

    /* get value */
    if (*p2) {
      p = p2;
      while (*p && isspace((int) *p)) p++;

      value = p;

      p2 = value + strlen(value);
      while (isspace((int) *(p2-1))) *--p2 = '\0';
    }
    else {
      value = 0;
    }

#define WTF ERR("%s: %d: wtf?", f, line);

    if (0) { }
    else if (strcasecmp(name, "alignment") == 0) {
      if (value) {
        if (strcasecmp(value, "top_left") == 0) text_alignment = TOP_LEFT;
        else if (strcasecmp(value, "top_right") == 0) text_alignment = TOP_RIGHT;
        else if (strcasecmp(value, "bottom_left") == 0) text_alignment = BOTTOM_LEFT;
        else if (strcasecmp(value, "bottom_right") == 0) text_alignment = BOTTOM_RIGHT;
        else WTF
      }
      else
        WTF
    }
    else if (strcasecmp(name, "background") == 0) {
      fork_to_background = string_to_bool(value);
    }
    else if (strcasecmp(name, "default_color") == 0) {
      if (value)
        default_fg_color = get_color(value);
      else
        WTF
    }
    else if (strcasecmp(name, "default_shadecolor") == 0) {
      if (value)
        default_bg_color = get_color(value);
      else
        WTF
    }
    else if (strcasecmp(name, "draw_borders") == 0) {
      draw_borders = string_to_bool(value);
    }
    else if (strcasecmp(name, "draw_shades") == 0) {
      draw_shades = string_to_bool(value);
    }
    else if (strcasecmp(name, "font") == 0) {
      if (value)
        font = strdup(value);
      else
        WTF
    }
    else if (strcasecmp(name, "gap_x") == 0) {
      if (value)
        gap_x = atoi(value);
      else
        WTF
    }
    else if (strcasecmp(name, "gap_y") == 0) {
      if (value)
        gap_y = atoi(value);
      else
        WTF
    }
    else if (strcasecmp(name, "no_buffers") == 0) {
      no_buffers = string_to_bool(value);
    }
    else if (strcasecmp(name, "stippled_borders") == 0) {
      if(value)
        stippled_borders = strtol(value, 0, 0);
      else
        stippled_borders = 4;
    }
    else if (strcasecmp(name, "temp1") == 0) {
      ERR("temp1 configuration is obsolete, use ${i2c <i2c device here> temp 1}");
    }
    else if (strcasecmp(name, "temp2") == 0) {
      ERR("temp2 configuration is obsolete, use ${i2c <i2c device here> temp 2}");
    }
    else if (strcasecmp(name, "update_interval") == 0) {
      if (value)
        update_interval = strtod(value, 0);
      else
        WTF
    }
    else if (strcasecmp(name, "uppercase") == 0) {
      stuff_in_upper_case = string_to_bool(value);
    }
    else if (strcasecmp(name, "text") == 0) {
      if (text != original_text)
        free(text);

      text = (char *) malloc(1);
      text[0] = '\0';

      while (!feof(fp)) {
        if (fgets(buf, 256, fp) == NULL) break;
        text = (char *) realloc(text, strlen(text) + strlen(buf) + 1);
        strcat(text, buf);

        if (strlen(text) > 1024*8) break;
      }

      fclose(fp);
      return;
    }
    else
      WTF
  }

  fclose(fp);
}

#define ATOM(a) XInternAtom(display, #a, False)

static Window find_window_to_draw() {
  Atom type;
  int format, i, j;
  unsigned long nitems, bytes;
  unsigned int n;
  Window root = RootWindow(display, screen);
  Window win = root;
  Window troot, parent, *children;
  unsigned char *vwin;

  /* Nautilus sets NAUTILUS_DESKTOP_WINDOW_ID property to root window */

  if (XGetWindowProperty(display, root, ATOM(NAUTILUS_DESKTOP_WINDOW_ID),
        0, 1, False, XA_WINDOW, &type, &format, &nitems, &bytes,
        &vwin) == Success && type == XA_WINDOW) {
    XWindowAttributes attrs;

    win = * (Window *) vwin;
    XFree(vwin);

    fprintf(stderr, "torsmo: drawing to nautilus desktop window\n");

    /* xpenguins way, search subwindows with same size as display */

    for (i=0; i<10; i++) {
      XQueryTree(display, win, &troot, &parent, &children, &n);
      for (j=0; j<(int)n; j++) {
        if (XGetWindowAttributes(display, children[j], &attrs)) {
          if (attrs.width == display_width && attrs.height == display_height) {
            win = children[j];
            break;
          }
        }
      }

      XFree(children);

      if (j == (int)n) break;
    }

    return win;
  }

  /* some window managers set __SWM_VROOT to some child of root window */

  XQueryTree(display, win, &troot, &parent, &children, &n);
  for (i=0; i<(int)n; i++) {
    if (XGetWindowProperty(display, children[i], ATOM(__SWM_VROOT),
          0, 1, False, XA_WINDOW, &type, &format, &nitems, &bytes,
          &vwin) == Success && type == XA_WINDOW) {
      win = * (Window *) vwin;
      XFree(vwin);
      XFree(children);
      fprintf(stderr, "torsmo: drawing to window from __SWM_VROOT property\n");
      return win;
    }
    XFree(vwin);
  }
  XFree(children);

  fprintf(stderr, "torsmo: drawing to root window\n");
  return RootWindow(display, screen);
}

static void init_X() {
  /* get display and window and stuff */
  if ((display=XOpenDisplay(0)) == NULL)
    CRIT_ERR("can't open display: %s", XDisplayName(0));

  screen = DefaultScreen(display);
  display_width = DisplayWidth(display, screen);
  display_height = DisplayHeight(display, screen);

  if (!win)
    win = find_window_to_draw();

  XSelectInput(display, win, ExposureMask);

  /* prepare GC */
  {
    XGCValues values;
    gc = XCreateGC(display, win, 0, &values);
  }

//  XMapWindow(display, win);

  default_bg_color = BlackPixel(display, screen);
  default_fg_color = WhitePixel(display, screen);
}

static const char *getopt_string = "vVdt:f:u:hc:w:";

int main(int argc, char **argv) {
  /* handle command line parameters that don't change configs */

  while (1) {
    int c = getopt(argc, argv, getopt_string);
    if (c == -1) break;

    switch (c) {
    case 'v':
    case 'V':
      printf("torsmo " VERSION " compiled " __DATE__ "\n");
      return 0;

    case 'c':
      /* if cconfig is set to a strdup of CONFIG_FILE, free it (even though
       * free() does the NULL check itself;), then load optarg value */
      if (cconfig) free(cconfig);
      cconfig = strdup(optarg);
      break;

    case 'h':
      printf(
"Usage: %s [OPTION]...\n"
"   -V            version\n"
"   -c FILE       load config file\n"
"   -d            daemonize, fork to background\n"
"   -f FONT       font to use\n"
"   -h            help\n"
"   -t TEXT       text to render\n"
"   -u SECS       update interval\n"
"   -w WIN_ID     window id to draw\n"
, argv[0]);
      return 0;

    case 'w':
      win = strtol(optarg, 0, 0);
      break;

    case '?':
      exit(EXIT_FAILURE);
    }
  }

  /* X stuff must be initialized here to get default colors */
  init_X();

  /* load cconfig or CONFIG_FILE */

#ifdef CONFIG_FILE
  if (!cconfig)
  {
    /* load default config file */
    char buf[256];

    variable_substitute(CONFIG_FILE, buf, 256);

    if (buf[0] != '\0')
      cconfig = strdup(buf);
  }
#endif

  if (cconfig) load_config_file(cconfig);

  optind = 0;

  /* handle other command line arguments */

  while (1) {
    int c = getopt(argc, argv, getopt_string);
    if (c == -1) break;

    switch (c) {
    case 'd':
      fork_to_background = 1;
      break;

    case 'f':
      font = optarg;
      break;

    case 't':
      if (text != original_text) free(text);
      text = strdup(optarg);
      convert_escapes(text);
      break;

    case 'u':
      update_interval = strtod(optarg, 0);
      break;

    case '?':
      exit(EXIT_FAILURE);
    }
  }

  /* load font */
  if ((font_info = XLoadQueryFont(display, font)) == NULL)
    CRIT_ERR("can't load font '%s'", font);

  XSetFont(display, gc, font_info->fid);

  update_uname();
  generate_text();
  generate_text();

  update_text_area();
  draw_stuff();

  /* fork */
  if (fork_to_background) {
    int ret = fork();
    switch (ret) {
    case -1:
      ERR("can't fork() to background: %s", strerror(errno));
      break;

    case 0:
      break;

    default:
      fprintf(stderr, "torsmo: forked to background, pid is %d\n", ret);
      exit(0);
      break;
    }
  }

  /* set SIGHUP, SIGINT and SIGTERM handler */
  {
    struct sigaction sa;

    sa.sa_handler = hup_handler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = SA_RESTART;
    if (sigaction(SIGHUP, &sa, NULL) != 0) {
      CRIT_ERR("can't set signal handler for SIGHUP: %s", strerror(errno));
    }

    sa.sa_handler = term_handler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = SA_RESTART;

    if (sigaction(SIGINT, &sa, NULL) != 0) {
      CRIT_ERR("can't set signal handler for SIGINT: %s", strerror(errno));
    }

    sa.sa_handler = term_handler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = SA_RESTART;

    if (sigaction(SIGTERM, &sa, NULL) != 0) {
      CRIT_ERR("can't set signal handler for SIGTERM: %s", strerror(errno));
    }
  }

  main_loop();

  return 0;
}
